use std::path::PathBuf;
use tool_helper::format_if_error;
use crate::config_reader::Directory;

use super::{CommonProperties, Configuration, Error, File};

mod attribute_names {
    pub const NAME:        &'static str = "name";
    pub const HIDDEN:      &'static str = "hidden";
    pub const PADDED_SIZE: &'static str = "padded_size";
}

pub fn parse(xml: String) -> Result<Configuration, Error> {
    let mut config = Configuration::new();
    let parser     = format_if_error!(roxmltree::Document::parse(xml.as_str()), "Parsing XML file failed with: {error_text}")?;
    let children   = parser.root().children();

    for node in children {
        if node.is_element() && node.tag_name().name() == "ISO_Project" {
            parse_iso_project(node, &mut config)?;
        }
    }

    Ok(config)
}

fn parse_iso_project(iso_project: roxmltree::Node, config: &mut Configuration) -> Result<(), Error> {
    for node in iso_project.children() {
        if node.is_element() {
            match node.tag_name().name() {
                "Description" => parse_description(node, config),
                "Track"       => parse_track(node, config)?,
                _ => ()
            }
        }
    }

    Ok(())
}

fn parse_description(description: roxmltree::Node, config: &mut Configuration) {
    for node in description.descendants() {
        if node.is_element() {
            match node.tag_name().name() {
                "Publisher" => config.publisher    = Some(String::from(node.text().unwrap_or_default())),
                "License"   => config.license_path = Some(PathBuf::from(node.text().unwrap_or_default())),
                _ => ()
            }
        }
    }
}

fn parse_track(track: roxmltree::Node, config: &mut Configuration) -> Result<(), Error> {
    fn parse_file(file: roxmltree::Node, is_hidden: bool) -> Result<File, Error> {
        Ok(File{
            common: read_common_properties(&file, is_hidden)?,
            path:   PathBuf::from(file.text().unwrap_or_default()),
        })
    }

    fn parse_file_system(cur_node: roxmltree::Node, root: &mut Directory, mut is_hidden: bool) -> Result<(), Error> {
        for node in cur_node.children() {
            if node.is_element() {
                match node.tag_name().name() {
                    "File"      => root.add_file(parse_file(node, is_hidden)?),
                    "Directory" => {
                        is_hidden |= parse_boolean_attribute(&node, attribute_names::HIDDEN)?;
                        let mut new_dir = Directory::new(node.attribute(attribute_names::NAME).unwrap_or_default(), is_hidden);

                        parse_file_system(node, &mut new_dir, is_hidden)?;
                        root.add_dir(new_dir);
                    },
                    _ => (),
                }
            }
        }

        Ok(())
    }

    parse_file_system(track, &mut config.root, false)
}

fn read_common_properties(xml: &roxmltree::Node, is_hidden: bool) -> Result<CommonProperties, Error> {
    Ok(CommonProperties{
        name:        String::from(xml.attribute(attribute_names::NAME).unwrap_or_default()), 
        is_hidden:   is_hidden | parse_boolean_attribute(&xml, attribute_names::HIDDEN)?,
        padded_size: read_padded_size(&xml)? 
    })
}

fn read_padded_size(xml: &roxmltree::Node) -> Result<Option<usize>, Error> {
    if let Some(padded_attr) = xml.attribute(attribute_names::PADDED_SIZE) {
        let padded_size = format_if_error!(padded_attr.parse::<usize>(), "Failed reading \"{}\" as padded size: {error_text}", padded_attr)?;
        Ok(Some(padded_size))
    }

    else {
        Ok(None)
    }
}

fn parse_boolean_attribute(xml: &roxmltree::Node, attribute_name: &str) -> Result<bool, Error> {
    if let Some(bool_str) = xml.attribute(attribute_name) {
        match bool_str {
            "true"  | "1" => Ok(true),
            "false" | "0" => Ok(false),
            _ => {
                return Err(Error::from_text(format!("Attribute \"{}\" expects either true,false or 1,0 as valid attributes - found {}", attribute_name, bool_str)));
            }
        }
    }
    
    else {
        Ok(false)
    }
}