pub use tool_helper::Error;

pub mod config_reader;
pub mod encoder;
pub mod file_writer;
pub mod types;

use tool_helper::{format_if_error, Output, read_file};
use types::{CDDesc, Directory, Properties};

pub type CalculateLBAFunction = fn(&mut types::CDDesc);

struct ContentDumpAlignment {
    name:    usize,
    lba:     usize,
    size:    usize,
    ex_size: usize,
}

const DEFAULT_CONTENT_ALIGNMENT:ContentDumpAlignment = ContentDumpAlignment{name: 24, lba: 8, size: 8, ex_size: 8};

pub fn process(config: config_reader::Configuration, calculate_lba: CalculateLBAFunction) -> Result<CDDesc, Error> {
    let mut cd_desc = parse_configuration(config)?;

    calculate_lba(&mut cd_desc);
    Ok(cd_desc)
}

pub fn dump_content(cd_desc: &CDDesc, mut out: Output) -> Result<(), Error> {
    const NO_NAME:&'static str = "<No Name>";
    const ARROW:&'static str   = "|==>";
    const INDENT_STEP:usize    = 4;

    fn get_visible_indicator(is_hidden: bool) -> char {
        if is_hidden {
            '-'
        }

        else {
            'o'
        }
    }

    fn write_file(out: &mut Output, indent: usize, file_name: String, properties: &Properties, file_lba: usize, file_size: usize, file_ex_size: usize) -> Result<(), Error> {
        Ok(writeln!(out, "{:>indent$}File: ({}) {:<name_align$} @{:<lba_align$} ={:<size_align$} >{:<ex_size_align$}", ARROW, get_visible_indicator(properties.is_hidden), file_name, file_lba, file_size, file_ex_size, 
        indent=indent + ARROW.len(), name_align=DEFAULT_CONTENT_ALIGNMENT.name, lba_align=DEFAULT_CONTENT_ALIGNMENT.lba, size_align=DEFAULT_CONTENT_ALIGNMENT.size, ex_size_align=DEFAULT_CONTENT_ALIGNMENT.ex_size)?)
    }

    fn write_dir(out: &mut Output, indent: usize, dir_name: &str, properties: &Properties, dir_lba: usize) -> Result<(), Error> {
        macro_rules! LBA_OUT {
            () => {
                "{:<lba_align$}"
            };
        }
        let is_hidden = properties.is_hidden;

        write!(out, "{:>indent$}Dir:  ({}) {:<name_align$} @", ARROW, get_visible_indicator(is_hidden), dir_name, 
        indent=indent + ARROW.len(), name_align=DEFAULT_CONTENT_ALIGNMENT.name)?;

        if is_hidden {
            Ok(writeln!(out, LBA_OUT!(), "<None>", lba_align=DEFAULT_CONTENT_ALIGNMENT.lba)?)
        }

        else {
            Ok(writeln!(out, LBA_OUT!(), dir_lba, lba_align=DEFAULT_CONTENT_ALIGNMENT.lba)?)
        }
    }

    fn write_intro(out: &mut Output) -> Result<(), Error> {
        writeln!(out, "{:>indent$}Type: ( ) {:<name_align$} @{:<lba_align$} ={:<size_align$} >{:<ex_size_align$}", "", "NAME", "LBA", "SIZE", "EXTENDED SIZE",
        indent=ARROW.len(), name_align=DEFAULT_CONTENT_ALIGNMENT.name, lba_align=DEFAULT_CONTENT_ALIGNMENT.lba, size_align=DEFAULT_CONTENT_ALIGNMENT.size, ex_size_align=DEFAULT_CONTENT_ALIGNMENT.ex_size)?;
        Ok(writeln!(out, "")?)
    }

    fn dump_dir(dir: &Directory, out: &mut Output, indent: usize) -> Result<(), Error> {
        for file in dir.file_iter() {
            let file         = file.borrow();
            let file_name    = file.name.as_string().unwrap_or(NO_NAME.to_owned());
            let properties   = &file.properties;
            let file_lba     = file.get_track_rel_lba();
            let file_size    = file.properties.get_real_size();
            let file_ex_size = file.get_extended_size();

            write_file(out, indent, file_name, properties, file_lba, file_size, file_ex_size)?;
        }

        for dir in dir.dir_iter() {
            let dir        = dir.borrow();
            let dir_name   = dir.name.as_str().unwrap_or(NO_NAME);
            let properties = dir.properties.borrow();
            let dir_lba    = dir.get_track_rel_lba();

            write_dir(out, indent, dir_name, &properties, dir_lba)?;
            dump_dir(&dir, out, indent + INDENT_STEP)?;
        }

        Ok(())
    }

    format_if_error!(write_intro(&mut out),                         "Writing content dump intro failed with: {error_text}")?;
    format_if_error!(dump_dir(&cd_desc.root.borrow(), &mut out, 0), "Creating content dump failed with: {error_text}")
}

fn parse_configuration(config: config_reader::Configuration) -> Result<CDDesc, Error> {
    fn parse_dir(dst_dir: &mut types::Directory, src_dir: config_reader::Directory) -> Result<(), Error> {
        for member in src_dir.into_iter() {
            match member {
                config_reader::DirMember::Directory(dir) => {
                    let mut new_dir = types::Directory::new(dir.name.as_str())?;

                    new_dir.properties.borrow_mut().is_hidden = dir.is_hidden;
                    parse_dir(&mut new_dir, dir)?;
                    dst_dir.add_dir(new_dir);
                },

                config_reader::DirMember::File(file)     => {
                    let mut desc_file = types::File::new_regular(file.common.name.as_str(), read_file(file.path)?)?;

                    desc_file.properties.padded_size_bytes = file.common.padded_size;
                    desc_file.properties.is_hidden         = file.common.is_hidden;
                    dst_dir.add_file(desc_file);
                },
            }
        }

        Ok(())
    }

    let cd_desc = CDDesc::new();

    if let Some(publisher) = config.publisher {
        cd_desc.pvd.borrow_mut().set_publisher(publisher);
    }

    if let Some(license_path) = config.license_path {
        cd_desc.system_area.borrow_mut().license_file_path = Some(license_path);
    }

    parse_dir(&mut cd_desc.root.borrow_mut(), config.root)?;
    Ok(cd_desc)
}
